from .flag import Flag


class FlagSet:
    def __init__(self):
        self.usage = lambda x: print(x)
        self.flags = {}
        self.set_flags = {}
        self.shorthands = {}
        self.args = []
        self.parsed = False
        self.args_len_at_dash = -1

    def add_flagset(self, flagset):
        self.flags.update(flagset.flags)

    def add_flag(self, typ, name, value, shorthand=None, usage=None):
        flag = Flag(typ=typ, name=name, value=typ(value), shorthand=shorthand, usage=usage, def_value=str(value))
        if name in self.flags:
            # TODO
            print('PANIC FLAG ALREADY EXISTS')
            return

        # TODO normalize
        self.flags[name] = flag

        if not shorthand:
            return

        # TODO
        if len(shorthand) > 1:
            print('SHORTHAND TOO LONG')
            return

        if shorthand in self.shorthands:
            print('SHORTHAND ALREADY EXISTS')
            return
        self.flags[shorthand] = flag

    def set_flag(self, name, value):
        # TODO normalize
        nname = name

        if nname not in self.flags:
            # TODO
            print(self.flags)
            print(nname)
            print('No such flag')
            return ValueError

        flag = self.flags[nname]

        err = flag.set_value(value)
        if err is not None:
            if flag.shorthand != '':
                flag_name = '-{}, --{}'.format(flag.shorthand, flag.name)
            else:
                flag_name = '--{}'.format(flag.name)
            print('invalid argument {} for {} flag: {}'.format(value, flag_name, err))
            return ValueError

        if not flag.changed:
            self.set_flags[nname] = flag
            flag.changed = True

        if flag.deprecated:
            # TODO
            print('Flag --{} has been deprecated, {}'.format(flag.name, flag.deprecated))
        return None

    def parse(self, args):
        self.parsed = True

        if len(args) <= 0:
            return None

        err = self.parse_args(args)

        if err is not None:
            # TODO error handling
            print("main parse error")
            exit(2)

        return None

    def parse_args(self, args):
        while len(args) > 0:
            s = args[0]
            args = args[1:]

            if len(s) == 0 or s[0] != '-' or len(s) == 1:
                self.args.append(s)
                self.args.extend(args)
                return None

            if s[1] == '-':
                if len(s) == 2:
                    self.args_len_at_dash = len(self.args)
                    self.args.extend(args)
                    break
                args, err = self.parse_long_args(s, args)
            else:
                args, err = self.parse_short_args(s, args)

            if err is not None:
                return err
        return None

    def parse_long_args(self, s, args):
        a = args
        name = s[2:]
        print(a, name)

        if len(name) == 0 or name[0] == '-' or name[0] == '=':
            print('bad flag syntax {}'.format(s))
            return a, ValueError()

        # TODO normalize
        split = name.split('=')
        name = split[0]

        if name not in self.flags:
            #TODO
            return a, ValueError()

        flag = self.flags[name]
        if len(split) == 2:
            value = split[1]
        elif flag.no_opt_def_val:
            value = flag.no_opt_def_val
        elif len(a) > 0:
            value = a[0]
            a = a[1:]
        else:
            print('flag needs an argument: {}'.format(s))
            return a, ValueError()

        err = self.set_flag(flag.name, value)
        if err is not None:
            print('ERROR parse long args')
        return a, err

    def parse_short_args(self, s, args):
        shorthands = s[1:]

        while len(shorthands) > 0:
            shorthands, a, err = self.parse_single_short_arg(shorthands, args)
            if err is not None:
                return a, err
        return a, err

    def parse_single_short_arg(self, shorthands, args):
        out_args = args

        out_shorts = shorthands[1:]
        c = shorthands[0]

        if c not in self.shorthands:
            # TODO
            return out_shorts, out_args, ValueError()

        flag = self.shorthands[c]

        if len(shorthands) > 2 and shorthands[1] == '=':
            value = shorthands[2:]
            out_shorts = ''
        elif flag.no_opt_def_val != '':
            value = flag.no_opt_def_val
        elif len(shorthands) > 1:
            value = shorthands[1:]
            out_shorts = ''
        elif len(args) > 0:
            value = args[0]
            out_args = args[1:]
        else:
            print("ERROR parse short arg")
            # TODO
            return out_shorts, out_args, ValueError()

        # TODO shorthand_deprecated

        err = self.set_flag(flag.name, value)
        if err is not None:
            print("ERROR parse short arg 2")
        return out_shorts, out_args, err



def args_minus_first(arguments, arg):
    if arg in arguments:
        i = arguments.index(arg)
        return arguments[:i] + arguments[i+1:]
    return arguments


def strip_flags(args, cmd):
    if len(args) == 0:
        return args

    cmd.merge_persistent_flags()

    commands = []
    #flags = cmd.flags

    while len(args) > 0:
        s = args[0]
        print(s)
        args = args[1:]

        if s == '--':
            break
        elif (s[:2] == '--' and '=' not in s) or (s[:1] == '-' and '=' not in s and len(s) == 2):
            if len(args) <= 1:
                break
            else:
                args = args[1:]
                continue
        elif s != '' and s[:1] != '-':
            commands.append(s)

    print(commands)
    return commands



class Flag:
    def __init__(
            self,
            typ,
            name,
            value,
            usage=None,
            def_value=None,
            shorthand=None,
            no_opt_def_val=None,
            changed=False,
            deprecated=False,
            hidden=False):
        self.typ = typ
        self.name = name
        self.shorthand = shorthand
        self.usage = usage
        self.value = value
        self.def_value = def_value
        self.no_opt_def_val = no_opt_def_val
        self.changed = changed
        self.deprecated = deprecated
        self.hidden = hidden

    def get_value(self):
        return self.value

    def set_value(self, value):
        try:
            if self.typ == bool:
                self.value = bool(value)
            elif self.typ == float:
                self.value = float(value)
            elif self.typ == int:
                self.value = int(value)
            else:
                self.value = str(value)
            return None
        except (ValueError, TypeError):
            # TODO
            print('ValueError Warning')
            self.value = str(value)
            return ValueError()



import sys
from .flag_set import FlagSet
from .args import args_minus_first, strip_flags


class Command:
    def __init__(
            self,
            use,
            short_desc=None,
            long_desc=None,
            run=None):
        self.use = use
        self.short_desc = short_desc
        self.long_desc = long_desc
        self.run = run
        self.args = None
        self.on_init_func = None
        self.commands = []
        self.parent = None
        self.flags = FlagSet()
        self.pflags = FlagSet()
        self.parents_pflags = FlagSet()

    def execute(self):
        if self.has_parent():
            return self.parent.execute()

        self.on_init_func()

        args = self.args

        if self.args is None:
            args = sys.argv[1:]

        cmd, flags, err = self.find(args)

        if cmd is not None:
            cmd.execute_cmd(flags)
        else:
            print('ERROR command does not exist')

    def execute_cmd(self, args):
        err = self.parse_flags(args)
        self.run(args)

    def has_parent(self):
        return self.parent is not None

    def find_next(self, next_cmd):
        for cmd in self.commands:
            if cmd.name() == next_cmd:
                return cmd

    def find(self, args):
        def inner_find(c, inner_args):
            args_wo_flags = strip_flags(args, self)
            if len(args_wo_flags) == 0:
                return c, inner_args

            next_cmd = args_wo_flags[0]

            cmd = c.find_next(next_cmd)
            if cmd is not None:
                return inner_find(cmd, args_minus_first(inner_args, next_cmd))
            return c, inner_args

        command_found, a = inner_find(self, args)

        return command_found, a, None

    def merge_persistent_flags(self):
        self.update_parents_pflags()
        self.flags.add_flagset(self.pflags)
        self.flags.add_flagset(self.parents_pflags)

    def update_parents_pflags(self):
        def add_pflags(parent):
            self.parents_pflags.add_flagset(parent.pflags)
        self.visit_parent(add_pflags)

    def visit_parent(self, func):
        if self.has_parent():
            func = func(self.parent)
            self.parent.visit_parent(func)

    def name(self):
        name = self.use
        i = name.find(' ')
        if i != -1:
            name = name[:i]
        return name

    def on_initialize(self, func):
        self.on_init_func = func

    def add_command(self, cmd):
        cmd.parent = self
        self.commands.append(cmd)

    def parse_flags(self, args):
        self.merge_persistent_flags()

        err = self.flags.parse(args)
